//----------------------------------------------
// uNikeTCP v1.0
// Copyright (C) 2005 by Rst7/CBSIE
//  ATMega8
//  10 
// SLIP  115200 8N1
//----------------------------------------------

#include "iom8.h"
#include "inavr.h"
#include "setjmp.h"
#include "pgmspace.h"
#ifdef DEBUG
#include "stdio.h"
#endif

#define EOP SREG_Bit6
#define _CARRY SREG_Bit0

volatile char TRM_CHAR; //  
//
typedef char __flash * STR;

//IP 
struct
{
 char MIP0;
 char MIP1;
 char MIP2;
 char MIP3;
};

//   CRC
unsigned int MIP_CRC;

#define WIN_SIZE 128

//  SEQ
unsigned long ISS;

typedef struct
{
 char dotx; // 
 char state; // 
 unsigned long RIP; // IP 
 //TCP 
 unsigned int lport; //  (  )
 unsigned int rport; // 
 unsigned long SEQ; //    
 unsigned long ACK; //  
 union
 {
  char CTL; //
  struct
  {
   char FIN:1,
        SYN:1,
        RST:1,
        PSH:1,
        _ACK:1,
        F5:1,
        F6:1,
        F7:1;
  };
 };
 unsigned int wsize; //  
 unsigned int INCACK; //   ACK  
 unsigned int INCSEQ; //   SEQ  
 unsigned int LENS; //-  
 unsigned int rxp; //  . 
 unsigned int rxf; //. . 
 unsigned int txp; //  
 unsigned int txf; //.  
 char txb[WIN_SIZE]; // 
 char rxb[WIN_SIZE]; // 
} T_TCB;

// - TCB
#define MAX_TCB	2

T_TCB TCBs[MAX_TCB];

//  TCB
#define CLOSED		0
#define LISTEN		1
#define SYN_SENT	2
#define SYN_RECIVED	3
#define ESTABLISHED	4
#define CLOSE		5
#define FIN_WAIT_1	6
#define FIN_WAIT_2	7
#define CLOSE_WAIT	8
#define CLOSING		9
#define LAST_ACK	10
#define	TIME_WAIT	11

char ICMP_dotx; //  ICMP  (  )
unsigned long ICMP_RIP;
// ICMP
struct
{
 char type;
 char code;
 unsigned int crc;
 unsigned int id;
 unsigned int seq;
 char data[32];
} ICMP_HDR;

//=====================================================================
jmp_buf main_task;
jmp_buf iprx_task;
jmp_buf iptx_task;
char iprx_rstack[8];
char iprx_cstack[32];
char iptx_rstack[8];
char iptx_cstack[32];

#pragma vector=USART_RXC_vect
__interrupt void rxint(void)
{
 if (!setjmp(main_task)) //  . 
 {
  longjmp(iprx_task,1); //   IPrecive
 }
}

#pragma vector=USART_UDRE_vect
__interrupt void txint(void)
{
 if (!setjmp(main_task)) //  . 
 {
  longjmp(iptx_task,1); //   IPtransmit
 }
}

#pragma vector=TIMER0_OVF_vect
__interrupt void delayint(void)
{
 if (!setjmp(main_task)) //  . 
 {
  longjmp(iptx_task,1); //   IPtransmit
 }
}

//------------------------------------------------------------------------
//  32 
//------------------------------------------------------------------------
#pragma optimize=no_inline
__z_x copy32(unsigned long *a, unsigned long *b)
{
 *a=*b;
}

//  = a-b
#pragma optimize=no_inline
__x_z unsigned int sub32(unsigned long *a, unsigned long *b)
{
  return(__reverse(*a)-__reverse(*b));
}

// a=a-b
#pragma optimize=no_inline
__z inc32(unsigned long *a, unsigned int i)
{
 *a=__reverse(__reverse(*a)+i);
}

#pragma optimize=no_inline
void NextISS(void)
{
 inc32(&ISS,1);
}

//-----------------------------------------------------------
// 
//-----------------------------------------------------------
char rxparity;
unsigned int rxcrc;

#pragma optimize=no_inline
void subrxcrc(unsigned int i)
{
 i=rxcrc-i;
 if (_CARRY) i--;
 rxcrc=i;
}

#pragma optimize=no_inline
void setrxcrc(unsigned int i)
{
 rxcrc=i;
}

#pragma optimize=no_inline
unsigned int testrxcrc(void)
{
 return(rxcrc);
}

char _sliprx(void)
{
 char c;
 NextISS();
 if (!setjmp(iprx_task))
 {
  UCSRB|=(1<<RXCIE);
  longjmp(main_task,1);
 }
 UCSRB&=~(1<<RXCIE);
 if ((c=UDR)==0xdb)
 {
  if (!setjmp(iprx_task))
  {
   UCSRB|=(1<<RXCIE);
   longjmp(main_task,1);
  }
  UCSRB&=~(1<<RXCIE);
  if ((c=UDR)==0xc0) goto L_RXC0;
  c=c==0xdc?0xc0:0xdb; // 
  goto L_RXDATA;
 }
 if (c==0xc0)
 {
  L_RXC0:
  EOP=1;
  rxparity=0;
  setrxcrc(0);
  return(0);
 }
 L_RXDATA:
 {
  EOP=0;
  unsigned int i=c;
  if (rxparity=~rxparity) i<<=8;
  subrxcrc(i);
  return(c);
 }
}

void _txb(char c)
{
 UDR=c;
 NextISS();
 if (!setjmp(iptx_task))
 {
  UCSRB|=(1<<UDRIE);
  longjmp(main_task,1);
 }
 UCSRB&=~(1<<UDRIE);
}

void _tcp_delay(char c)
{
// ISS++;
 if (!setjmp(iptx_task))
 {
  TCNT0=c;
  TIFR=1<<TOV0;
  TIMSK|=(1<<TOV0);
  longjmp(main_task,1);
 }
 TIMSK&=~(1<<TOV0);
}

void _sliptx(char c)
{
 if(c==0xc0)
 {
  _txb(0xDB);
  c=0xDC;
 }
 else if (c==0xdb)
 {
  _txb(c);
  c=0xDD;
 }
 _txb(c);
}

//------------------------------------------------------------------------
// IP     
//------------------------------------------------------------------------

void ICMPrx(unsigned long rip, unsigned int ip_len);
void TCPrx(unsigned long rip, unsigned int ip_len);

#pragma optimize=no_inline
void TCPrx_jmp(unsigned long rip, unsigned int ip_len) {TCPrx(rip,ip_len);}
#pragma optimize=no_inline
void ICMPrx_jmp(unsigned long rip, unsigned int ip_len) {ICMPrx(rip,ip_len);}

__task __noreturn IPreciver(void)
{
 char ip_pos;
 char ip_hlen;
 char prtc;
 char c;
 unsigned int ip_len;
 unsigned long rip;

 //  
 do
 {
  TRM_CHAR=_sliprx();
 }
 while(!EOP);
 L_SLIP_START:
 ip_pos=0;
 ip_hlen=0;
 do
 {
  c=_sliprx();
  if (EOP) goto L_SLIP_START;
  switch(ip_pos)
  {
   case 0: ip_hlen=c<<2&0x3F; break;
   case 7: if (c) goto L_SLIP_END; break;
   case 2: ip_len=c<<8; break;
   case 3: ip_len|=c; break;
   case 9: prtc=c; break;
   case 12: rip=(unsigned long)c<<24; break;
   case 13: rip|=(unsigned long)c<<16; break;
   case 14: rip|=(unsigned long)c<<8; break;
   case 15: rip|=c; break;
   case 16: if (c!=MIP0) goto L_SLIP_END; break;
   case 17: if (c!=MIP1) goto L_SLIP_END; break;
   case 18: if (c!=MIP2) goto L_SLIP_END; break;
   case 19: if (c!=MIP3) goto L_SLIP_END; break;
  }
  ip_pos++;
 }
 while(ip_pos!=ip_hlen);
 if (testrxcrc()) goto L_SLIP_END; //   
 ip_len-=ip_hlen; // IP 
 switch(prtc)
 {
  case 1:
   ICMPrx_jmp(rip,ip_len);
   break;
  case 6:
   TCPrx_jmp(rip,ip_len);
   break;
 }
 L_SLIP_END:
 do
 {
   _sliprx();
 }
 while(!EOP);
 goto L_SLIP_START;
}

// IP    6 ( TCP   )
//   IP   
__noreturn void TCPrx(unsigned long rip, unsigned int iplen)
{
 char c;
 char n;
 unsigned int i;
 int tcp_pos;
 //TCP 
 struct
 {
  unsigned int rport; //  
  unsigned int lport; // 
  unsigned long SEQ;
  unsigned long ACK;
  char hlen;
  union
  {
   char CTL; //
   struct
   {
    char FIN:1,
         SYN:1,
         RST:1,
         PSH:1,
         _ACK:1,
         F5:1,
         F6:1,
         F7:1;
   };
  };
  unsigned int wsize;
 } TCP_HDR;

 T_TCB *tcb=TCBs;
 // CRC 
 setrxcrc(MIP_CRC);
 subrxcrc(iplen+6); // !!!!
 subrxcrc(rip>>16);
 subrxcrc(rip);

 i=0xFFFF;
 TCP_HDR.hlen=20;
 do
 {
  i++;
  c=_sliprx();
  if (EOP) return;
  if (i==12) c=c>>2&0x3C; //  
  if (i<sizeof(TCP_HDR))
  {
   // 
   ((char *)&TCP_HDR)[i]=c;
   continue;
  }
  tcp_pos=i-TCP_HDR.hlen;
  if (tcp_pos==-1) //  
  {
   //  
   //  TCB
   n=MAX_TCB;
   do
   {
    if (
        (
         (((unsigned long)TCP_HDR.rport<<16)|TCP_HDR.lport)
         ==
         *((unsigned long *)&tcb->lport)
        )&&
//        (TCP_HDR.rport==tcb->rport)&&
//        (TCP_HDR.lport==tcb->lport)&&
        (rip==__reverse(tcb->RIP))
       )
    break;
    tcb++;
   }
   while(--n);
  }
  if (tcp_pos<0) continue; //  
  if (n&&tcp_pos<tcb->rxf) tcb->rxb[(tcb->rxp+tcp_pos)&(WIN_SIZE-1)]=c;
 }
 while(--iplen);
 if (testrxcrc()) return; //   
 if (n==0) //   TCB
 {
  if (TCP_HDR.CTL!=2) return; //    
  n=MAX_TCB;
  do
  {
   //  TCB
   tcb--;
   if ((tcb->state==LISTEN)&&(tcb->lport==TCP_HDR.lport))
   {
    //  c  
    tcb->RIP=__reverse(rip);
    tcb->state=SYN_RECIVED;
    tcb->rport=TCP_HDR.rport;
    copy32(&tcb->SEQ,&ISS);
//    tcb->SEQ=__reverse(ISS);
    copy32(&tcb->ACK,&TCP_HDR.SEQ);
    inc32(&tcb->ACK,1);
//    tcb->ACK=__reverse(__reverse(TCP_HDR.SEQ)+1);
    tcb->CTL=0x12; //SYN=1;ACK=1;
    tcb->dotx=1; // 
    return;
   }
  }
  while(--n);
  return; //  TCB
 }
 //  TCP
 tcb->wsize=TCP_HDR.wsize; // 
 if (TCP_HDR.RST)
 {
  tcb->state=CLOSED;
  tcb->CTL=0;
  return;
 }
 switch(tcb->state)
 {
  case CLOSED:
   return;
  case SYN_RECIVED:
   if (TCP_HDR.SYN) return;
   if (TCP_HDR._ACK)
   {
    inc32(&tcb->SEQ,1);
    if (!sub32(&tcb->SEQ,&TCP_HDR.ACK))
    {
     tcb->state=ESTABLISHED;
     tcb->CTL=0x10; //_ACK=1;
    }
    else
    {
     tcb->CTL=0x04; //RST=1;
     //tcb->SEQ=TCP_HDR.ACK;
     copy32(&tcb->SEQ,&TCP_HDR.ACK);
    }
    tcb->dotx=1;
    return;
   }
   return;
  case FIN_WAIT_1:
  case CLOSE_WAIT:
  case CLOSE:
  case ESTABLISHED:
   i=sub32(&TCP_HDR.ACK,&tcb->SEQ);
   if (tcb->FIN&&(i==1))
   {
     // FIN   
     inc32(&tcb->SEQ,1); //- FIN  
     tcb->FIN=0;
     if (tcb->state==CLOSE) tcb->state=FIN_WAIT_1; // FIN  
     i=0;
   }
   if (i<=tcb->LENS) tcb->INCSEQ=i; //     
   //,   
   tcp_pos++; // tcp_pos   
   i=0;
   if (!sub32(&TCP_HDR.SEQ,&tcb->ACK))
   {
    //  ,  
    if (tcp_pos<=tcb->rxf) //    
     i=tcp_pos; //  .  
    else
     i=tcb->rxf; //    
    tcb->rxp=(tcb->rxp+i)&(WIN_SIZE-1); //    
    tcb->rxf-=i; // -  
    if (TCP_HDR.FIN) // FIN
    {
     i++; //  FIN
     switch(tcb->state)
     {
     case ESTABLISHED:
      tcb->state=CLOSE_WAIT; // FIN c  ,  ACK  FIN (  )
      break;
     case CLOSE:
     case FIN_WAIT_1:
      tcb->state=CLOSING; // FIN   CLOSE,  ACK   
      break;
     }
    }
    else
    {
     if (tcb->state==FIN_WAIT_1) return; //   FIN  CLOSE;
     if (tcb->state==CLOSE_WAIT) tcb->state=CLOSED; // ACK  ACK  FIN,  
    }
   }
   tcb->INCACK=i;
   tcb->dotx=1;
   return;
 }
}

// ICMP    1 ( ICMP   )
//   IP   
__noreturn void ICMPrx(unsigned long rip, unsigned int iplen)
{
 unsigned int i;
 char *p=&ICMP_HDR.type;
 char c;
 setrxcrc(0);
 i=0;
 do
 {
  c=_sliprx();
  if (EOP) return;
  if (i<sizeof(ICMP_HDR)) *p++=c;
  i++;
 }
 while(--iplen);
 if (testrxcrc()) return; //  . 
 if (ICMP_HDR.type!=8) return; // 
 ICMP_HDR.type=0; //
 if (i>sizeof(ICMP_HDR)) i=sizeof(ICMP_HDR);
 ICMP_RIP=__reverse(rip);
 ICMP_dotx=i;
}

//----------------------------------------------------------------
// 
//----------------------------------------------------------------
unsigned int txcrc;

#pragma optimize=no_inline
void subtxcrc(unsigned int i)
{
 i=txcrc-i;
 if (_CARRY) i--;
 txcrc=i;
}

#pragma optimize=no_inline
void settxcrc(unsigned int i)
{
 txcrc=i;
}

#pragma optimize=no_inline
unsigned int gettxcrc(void)
{
 return(txcrc);
}

#pragma optimize=no_inline
void _sliptx4i(unsigned long i)
{
 _sliptx(i>>24);
 _sliptx(i>>16);
 subtxcrc(i>>16);
 _sliptx(i>>8);
 _sliptx(i);
 subtxcrc(i);
}

#pragma optimize=no_inline
__z void _sliptx4p(unsigned long *p)
{
 _sliptx4i(__reverse(*p));
}

__x void IPHDRtransmit(char prot, unsigned int ip_len, unsigned long *ipadr)
{
  // CRC IP 
  ip_len+=20; // + 
  settxcrc(MIP_CRC);
  subtxcrc(__reverse(*ipadr)>>16);
  subtxcrc(__reverse(*ipadr));
  subtxcrc(0x8000|prot);
  _txb(0xc0); //  SLIP
  _sliptx4i(0x45000000L|ip_len); //    / TOS /   
  _sliptx4i(0x00004000L); //Indentification / Flag+fragment offset
  _sliptx4i(0x80000000L|((unsigned long)prot<<16)|gettxcrc()); //TTL/Protocol/ 
  settxcrc((unsigned int)-7); //  CR  TCP
  _sliptx4p((unsigned long *)&MIP0);
  _sliptx4p(ipadr);
  // txcrc - CRC TCP-  tcplen
}

//   IP 
__task  __noreturn void IPtransmiter(void)
{
 char c;
 char n;
 unsigned int ip_len;
 unsigned int ip_pos;
 unsigned int tcp_pos;
 T_TCB *tcb;
 for(;;)
 {
  if ((c=TRM_CHAR)!=0)
  {
   _txb(c);
   if (c=='T') // 
   {
    _txb('S');
    _txb('E');
    _txb('R');
    _txb('V');
    _txb('E');
    _txb('R');
   }
   TRM_CHAR=0;
  }
  if ((ip_len=ICMP_dotx)!=0)
  {
   char *p;
   //   ICMP 
   ICMP_dotx=0; //   . 
   IPHDRtransmit(1,ip_len,&ICMP_RIP);
   settxcrc(ICMP_HDR.crc=0);
   p=(char *)&ICMP_HDR;
   ip_pos=0;
   while(ip_pos<ip_len)
   {
    c=*p++;
    if (ip_pos&1)
     subtxcrc(c);
    else
     subtxcrc(c<<8);
    ip_pos++;
   }
   ICMP_HDR.crc=__reverse(gettxcrc());
   p=(char *)&ICMP_HDR;
   ip_pos=0;
   while(ip_pos<ip_len)
   {
    _sliptx(*p++);
    ip_pos++;
   }
   _txb(0xc0);
  }
  tcb=TCBs;
  n=0;
  do
  {
   ip_len=0;
   tcp_pos=tcb->txf;
   c=tcb->state;
   if (c<SYN_SENT) goto L_NEXT_TCB; //  
   //if (tcb->dotx==0) goto L_NEXT_TCB; //  = 0
   if (--tcb->dotx) goto L_NEXT_TCB; //,      0, . 
   //L_DO_TCB:
   //    
   inc32(&tcb->ACK,tcb->INCACK); //   
   if (c==CLOSE_WAIT) //  
   {
    tcb->FIN=1; // FIN   FIN
   }
   else tcb->FIN=0;
   if (c==ESTABLISHED||c==CLOSE) //  
   {
    ip_len=tcb->INCSEQ;
    //  ,   
    tcb->txp=(tcb->txp+ip_len)&(WIN_SIZE-1); //     - 
    tcp_pos-=ip_len; // -     - 
    inc32(&tcb->SEQ,ip_len); //   
    //  
    tcb->txf=ip_len=tcp_pos; //-   
    ip_pos=__reverse(tcb->wsize); // 
    if (ip_len>ip_pos) ip_len=ip_pos; // -   ,    
    tcb->INCSEQ=0;
    //   
    if((ip_len==0)&&(tcb->INCACK==0)&&(c==ESTABLISHED))
    {
     tcb->LENS=0;
     goto L_NEXT_TCB;
    }
    //================================
    if (c==CLOSE&&(!ip_len))
    {
     tcb->FIN=1; //    ,    
    }
   }
   // 
   tcb->INCACK=0;
   tcb->LENS=ip_len; //   
   // 
   ip_len+=20; // +  TCP
   IPHDRtransmit(6,ip_len,&tcb->RIP); //IP 
   // TCP  (20 )
   // RIP MIP tcplen prot -  tcplen,   IPHDRtransmit
   //lprt rprt SEQH SEQL ACKH ACKL 50ff WSIZ CRC URG
   subtxcrc(ip_len); //  TCP   CRC
   _sliptx4p((unsigned long *)&tcb->lport);
   _sliptx4p(&tcb->SEQ);
   _sliptx4p(&tcb->ACK);
   _sliptx4i(0x50000000L| //Header length
             ((unsigned long)(tcb->CTL&0x1F)<<16)| //Flags
             tcb->rxf); //Window size
   // .  
   tcp_pos=tcb->txp;
   ip_pos=20;
   while(ip_pos<ip_len)
   {
    c=tcb->txb[tcp_pos];
    tcp_pos++;
    tcp_pos&=WIN_SIZE-1;
    if (ip_pos&1)
      subtxcrc(c);
    else
      subtxcrc(c<<8);
    ip_pos++;
   }
   _sliptx4i((unsigned long)gettxcrc()<<16); //CRC & Urgent pointer
   tcp_pos=tcb->txp;
   ip_pos=20;
   while(ip_pos<ip_len)
   {
    _sliptx(tcb->txb[tcp_pos]);
    tcp_pos++;
    tcp_pos&=WIN_SIZE-1;
    ip_pos++;
   }
   _txb(0xc0);
   if (tcb->state==CLOSING) tcb->state=CLOSED; // ACK  FIN
   L_NEXT_TCB:
   tcb++;
   n++;
  }
  while(n<MAX_TCB);
  _tcp_delay(178);
 }
}


//=================================================================================
void sys_init(void)
{
 PORTD=0x03;
 DDRD=0xFE;
 DDRB=0x00;
 PORTB=0xFF;
 DDRC=0x3F;
 UCSRB=(1<<TXEN)+(1<<RXEN);
 UCSRA=0;
 UBRRL=5; //115200
 TCCR0=5; // 1:1024
}
//=================================================================================
void TCP_init(char ip0, char ip1, char ip2, char ip3)
{
 MIP0=ip0;
 MIP1=ip1;
 MIP2=ip2;
 MIP3=ip3;
 setrxcrc(0);
 subrxcrc((unsigned int)((ip0<<8)|ip1));
 subrxcrc((unsigned int)((ip2<<8)|ip3));
 MIP_CRC=testrxcrc();
 //  TCPIP 
 __disable_interrupt();
 ((unsigned int *)iprx_task)[10]=((unsigned int)iprx_rstack)+7; //SP
 ((unsigned int *)iprx_task)[8]=((unsigned int)iprx_cstack)+32; //Y
 ((unsigned int *)iprx_task)[9]=(unsigned int)IPreciver; // 
 ((unsigned int *)iptx_task)[10]=((unsigned int)iptx_rstack)+7; //SP
 ((unsigned int *)iptx_task)[8]=((unsigned int)iptx_cstack)+32; //Y
 ((unsigned int *)iptx_task)[9]=(unsigned int)IPtransmiter; // 
 if (!setjmp(main_task)) longjmp(iptx_task,1); //  IPtransmit
 if (!setjmp(main_task)) longjmp(iprx_task,1); //  IPrecive
}

//    
T_TCB *TCPlisten(unsigned int port)
{
 T_TCB *tcb=TCBs;
 char i=0;
 do
 {
  __disable_interrupt();
  if (tcb->state==CLOSED)
  {
   tcb->state=LISTEN;
   tcb->lport=__reverse(port);
   tcb->rxp=0;
   tcb->rxf=WIN_SIZE;
   tcb->txp=0;
   tcb->txf=0;
   tcb->CTL=0;
   tcb->INCACK=0;
   tcb->INCSEQ=0;
   __enable_interrupt();
   return(tcb);
  }
  __enable_interrupt();
  tcb++;
  i++;
 }
 while(i<MAX_TCB);
 return(0);
}

// 
__z void TCPclose(T_TCB *tcb)
{
 __disable_interrupt();
 if (tcb->state==ESTABLISHED)
 {
  tcb->state=CLOSE;
  tcb->dotx=1;
 }
 else
  tcb->state=CLOSED;
 __enable_interrupt();
}

__z int TCPgetc(T_TCB *tcb)
{
 char c;
 if (tcb==0) return -1;
 if (tcb->state!=ESTABLISHED) return -1;
 while(tcb->rxf==WIN_SIZE); //   
 __disable_interrupt();
 c=tcb->rxb[(tcb->rxp+tcb->rxf)&(WIN_SIZE-1)];
 tcb->rxf++;
 __enable_interrupt();
 return c;
}

__z void TCPputc(T_TCB * tcb, char c)
{
 if (tcb==0) return;
 if (tcb->state!=ESTABLISHED) return;
 while(tcb->txf==WIN_SIZE)
 {
  __disable_interrupt();
  if (tcb->INCSEQ) tcb->dotx=1; //   ,  
  __enable_interrupt();
 }; //  
 __disable_interrupt();
 tcb->txb[(tcb->txp+tcb->txf)&(WIN_SIZE-1)]=c;
 tcb->txf++;
 if (tcb->txf==WIN_SIZE) tcb->dotx=1; //     20
 __enable_interrupt();
}

/*__z_x int TCPgets(T_TCB * tcb, char *p, char maxlen)
{
 int c;
 char i=0;
 for(;;)
 {
  c=TCPgetc(tcb);
  if (c<0) return c;
  if (c==0x0a) continue;
  if (c==0x0d) break;
  if (i<maxlen) *p++=c;
  i++;
 }
 *p=0;
 return(i);
}*/

__x_z void TCPputs_P(T_TCB * tcb,STR p)
{
 char c;
 while((c=*p++)!=0) TCPputc(tcb,c);
}
//===========================================================================
//===========================================================================
//===========================================================================

#define TELNET_PORT 23
void TELNETdaemon(void)
{
 static T_TCB *st_sock=0;
 T_TCB *sock;
 static char state=0;
 char c;

 sock=st_sock;
 if (state)
 {
  if (sock->state==CLOSED) state=0;
 }
 switch(state)
 {
  case 0:
   sock=TCPlisten(TELNET_PORT);
   if (sock!=0)
   {
     st_sock=sock;
     state=1;
   }
   break;
  case 1:
   if (sock->state==ESTABLISHED)
   {
    TCPputs_P(sock,"Wellcome to MicroTelnetServer v0.0\r\nCopyright (C) 2005 by Rst7/CBSIE\r\n");
    state=2;
   }
   break;
  case 2:
   if (sock->rxf!=WIN_SIZE)
   {
    c=TCPgetc(sock);
    if (c==0x1B)
    {
     TCPclose(sock);
     state=0;
     break;
    }
    if (c>0x40&&c<0x7F) c^=0x20;
    TCPputc(sock,c);
   }
   break;
 }
}

#define HTTP_PORT 80
const __flash char INDEX_HTML[] @ "FILESYS";

unsigned int rint(char *p)
{
 unsigned int i=0;
 char c;
 for(;;)
 {
  c=*p++;
  c-='0';
  if(SREG&1) break;
  if(c>9) break;
  i*=10;
  i+=c;
 }
 return(i);
}

void HTTPdaemon(void)
{
 static T_TCB *st_sock=0;
 T_TCB *sock;
 static char state=0;
 static char s_req[16];
 static char s[32];
 char c;
 static unsigned int cev;
 static char oldpinb;
 static char newpinb;
 unsigned int rev;

 sock=st_sock;
 if (state)
 {
  if (sock->state==CLOSED) state=0;
 }
 if ((state>1)&&(state<34))
 {
  //  
  if (sock->rxf!=WIN_SIZE)
  {
   c=TCPgetc(sock);
   if (c==0x0a) return;
   if (c==0x0d)
   {
    s[state-2]=0;
    state=35;
   }
   else
   {
    s[state-2]=c;
    state++;
   }
  }
 }
 switch(state)
 {
  case 0:
   sock=TCPlisten(HTTP_PORT);
   if (sock!=0)
   {
     st_sock=sock;
     state=1;
   }
   break;
  case 1:
   if (sock->state==ESTABLISHED)
   {
    state=2;
   }
   break;
  case 34:
   //  
   if (sock->rxf!=WIN_SIZE)
   {
    if (TCPgetc(sock)==0x0d) state=35;
   }
   break;
  case 35:
   if (strncmp_P(s,(PGM_P)"GET /",5)==0) memcpy(s_req,s+5,16);
   if (s[0]!=0)
   {
    state=2; //  
    break;
   }
   switch(s_req[0])
   {
    case ' ':
     cev=0;
     oldpinb=0;
     TCPputs_P(sock,(STR)"HTTP/1.0 200 OK\r\n\r\n");
     TCPputs_P(sock,(STR)"<HTML><HEAD><TITLE>  </TITLE>");
     TCPputs_P(sock,(STR)"<META http-equiv=Content-Type content=\"text/html; charset=windows-1251\">");
     TCPputs_P(sock,(STR)"</HEAD><FRAMESET ROWS=\"60,*\"><FRAME NAME=\"rxevent\" NORESIZE>");
     TCPputs_P(sock,(STR)"<FRAME SRC=2.htm NAME=\"view\"></FRAMESET></HTML>");
     break;
    case '2':
     TCPputs_P(sock,(STR)"HTTP/1.0 200 OK\r\n\r\n");
     TCPputs_P(sock,(STR)"<HTML><HEAD>");
     TCPputs_P(sock,(STR)"<META http-equiv=Content-Type content=\"text/html; charset=windows-1251\">");
     TCPputs_P(sock,(STR)"<SCRIPT LANGUAGE=\"JavaScript\">var REQms;var IDtimeout=-1;var curevent=0;");
     TCPputs_P(sock,(STR)"function updlog(){if(IDtimeout>0)clearTimeout(IDtimeout);d=new Date();");
     TCPputs_P(sock,(STR)"REQms=d.getTime();doreq.action=\"event\"+curevent+\".htm\";doreq.submit();");
     TCPputs_P(sock,(STR)"IDtimeout=setTimeout(\"noconnect()\",30000);}");
     TCPputs_P(sock,(STR)"function noconnect(){alert(\"      30 !\");");
     TCPputs_P(sock,(STR)"updlog();}function Refresh0(){curevent=0;document.term.log_field.value=\"\";");
     TCPputs_P(sock,(STR)"updlog();}</SCRIPT></HEAD><BODY onLoad=\"updlog()\">");
     TCPputs_P(sock,(STR)"<H1>  ...</H1><FORM NAME=term>");
     TCPputs_P(sock,(STR)"<TEXTAREA NAME=log_field COLS=80 ROWS=20 WRAP=ON></TEXTAREA>");
     TCPputs_P(sock,(STR)"<INPUT TYPE=\"button\" NAME=\"from0\" VALUE=\"  \" ");
     TCPputs_P(sock,(STR)"onClick=\"Refresh0()\"></FORM><FORM NAME=\"doreq\" TARGET=\"rxevent\" ");
     TCPputs_P(sock,(STR)"METHOD=GET></FORM></BODY></HTML>");
     break;
    case 'e':
     rev=rint(s_req+5);
     if (!rev)
     {
      cev=0;
      oldpinb=0;
      newpinb=PINB;
     }
     if (rev==(cev+1))
     {
      //  
      cev++;
      oldpinb=newpinb;
     }
     else newpinb=PINB;
     if (rev!=cev) goto L_HTTP_NOT_FOUND;
     TCPputs_P(sock,(STR)"HTTP/1.0 200 OK\r\n\r\n");
     TCPputs_P(sock,(STR)"<HTML>");
     TCPputs_P(sock,(STR)"<META http-equiv=Content-Type content=\"text/html; charset=windows-1251\">");
     TCPputs_P(sock,(STR)"<SCRIPT LANGUAGE=\"JavaScript\">function x(){var t;");
     //  
     if (newpinb!=oldpinb)
     {
      // 
      TCPputs_P(sock,(STR)"parent.view.document.term.log_field.value+=(");
      TCPputs_P(sock,(STR)"\"  PINB -> ");
      c=0x80;
      do
      {
       TCPputc(sock,c&newpinb?'1':'0');
      }
      while(c>>=1);
      TCPputs_P(sock,(STR)"\\n\");");
      TCPputs_P(sock,(STR)"parent.view.curevent++;");
     }
     TCPputs_P(sock,(STR)"t=(new Date).getTime()-parent.view.REQms;document.write(\"  \"+t+\"\\n\");");
     TCPputs_P(sock,(STR)"parent.view.setTimeout(\"updlog()\",t+");
     if (newpinb!=oldpinb)
     {
      TCPputs_P(sock,(STR)"100");
     }
     else
     {
      TCPputs_P(sock,(STR)"3000");
     }
     TCPputs_P(sock,(STR)");}</SCRIPT><BODY onLoad=\"x()\"></BODY></HTML>");
     break;
    default:
     L_HTTP_NOT_FOUND:
     TCPputs_P(sock,(STR)"HTTP/1.0 404 Not found\r\n\r\n");
     {
      char *p=s_req;
      TCPputc(sock,rev>>8);
      TCPputc(sock,(char)rev);
      while((c=*p++)!=0)
      {

       if (p>(s_req+15)) break;
       TCPputc(sock,c);
      }
     }
     break;
   }
   TCPclose(sock);
   state=0;
   break;
 }
}

__task  __noreturn void main(void)
{
 sys_init();
 TCP_init(192,168,2,2);
 __enable_interrupt();
 #define CTS	PORTD_Bit2
 CTS=1;
 for(;;)
 {
  TELNETdaemon();
  HTTPdaemon();
 }
}

